;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2009  Luca Saiu
;;; Updated in 2010 by Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;;; This file provides an implementation of the game state, used in the
;;; server.


;;; The module we include here is an implementation of records for Guile.
;;; It's explained very well in the manual:
(use-modules (srfi srfi-9))

;;; We need some constants defined in this module:
(load "big-project-demo-rules.scm")

;;; Note that there are several global variables in this source file
;;; which are used in an imperative way, and we need to protect them
;;; from concurrent access with a mutex. Such synchronization is *not*
;;; implemented here, and it's the caller's responsibility.

;;; We represent the current state of each player (we can have arbitrarily
;;; many players in this particular game) as a record:
(define-record-type player
  (make-player-by-specifying-all-fields
      name
      red green blue
      score
      x
      y
      angle ; in radiants
      translation-state ; one of: forward, backward, idle
      rotation-state ; one of: left, right, idle
      update-frequency ; how many updates per second the client wants
      changed-update-frequency) ; has the update frequency just changed?
  player?
  (name              player-name              player-set-name!)
  (red               player-red               player-set-red!)
  (green             player-green             player-set-green!)
  (blue              player-blue              player-set-blue!)
  (score             player-score             player-set-score!)
  (x                 player-x                 player-set-x!)
  (y                 player-y                 player-set-y!)
  (angle             player-angle             player-set-angle!)
  (translation-state player-translation-state player-set-translation-state!)
  (rotation-state    player-rotation-state    player-set-rotation-state!)
  (update-frequency  player-update-frequency  player-set-update-frequency!)
  (changed-update-frequency  player-changed-update-frequency player-set-changed-update-frequency!))

;;; A procedure for making a player without specifying a value for all the 
;;; fields:
(define (make-player name)
  (make-player-by-specifying-all-fields name
                               (random 1.0) (random 1.0) (random 1.0)
                               0
                               (random-x-in-the-field)
                               (random-y-in-the-field)
                               (normalize-angle (random 2pi))
                               'idle
                               'idle
                               default-server-update-sending-frequency
                               #f))

;;; Add a new player to the game:
(define (add-player! player)
  (set! players (cons player players)))

;;; Remove a player from the game:
(define (remove-player! player)
  (set! players (without-by-identity player players)))

;;; Add the given number of new pills:
(define (add-pills! how-many)
  (set! pills (append (map (lambda whatever
                             (make-pill))
                           (iota how-many))
                      pills)))

;;; Remove an eaten pill:
(define (remove-pill! pill)
  (set! pills (without-by-identity pill pills)))

;;; Compute the next game state, imperatively set the global variables and
;;; return a message with state changes:
(define (update-state! seconds-since-the-last-update)
  (update-pill-state! seconds-since-the-last-update)
  (iterate (lambda (player)
             (update-player-state! player seconds-since-the-last-update))
           players)
  ;; Return the update message:
  (list  'updated-state
         (map (lambda (p)
                (list (player-name p) (player-score p)
                      (player-x p) (player-y p) (player-angle p)
                      (player-red p) (player-green p) (player-blue p)))
              players)
         (map (lambda (pill)
                (cons (pill-x pill) (pill-y pill)))
              pills)))

;;; Update the state of all the currently-existing pills:
(define (update-pill-state! delta-t) ; delta-t is in seconds
  (iterate (lambda (pill)
             (let* ((new-x (+ (pill-x pill)
                              (* (pill-velocity-x pill) delta-t)))
                    (new-y (+ (pill-y pill)
                              (* (pill-velocity-y pill) delta-t)))
                    (bounced-x (not (in-field-x? new-x)))
                    (bounced-y (not (in-field-y? new-y))))
               (pill-set-x! pill (saturate-x new-x))
               (pill-set-y! pill (saturate-y new-y))
               (when bounced-x
                 (pill-set-velocity-x! pill (- (pill-velocity-x pill))))
               (when bounced-y
                 (pill-set-velocity-y! pill (- (pill-velocity-y pill))))))
           pills))

;;; Update the state of a player:
(define (update-player-state! player delta-t) ; delta-t is in seconds
  ;; Update angle, x and y according to the current states:
  (let* ((old-angle (player-angle player))
         (old-x (player-x player))
         (old-y (player-y player))
         (rotation-quotient (case (player-rotation-state player)
                              ((left) 1)
                              ((right) -1)
                              ((idle) 0)))
         (translation-quotient (case (player-translation-state player)
                                 ((forward) 1)
                                 ((backward) -1)
                                 ((idle) 0)))
         (new-angle (normalize-angle (+ old-angle
                                        (* rotation-quotient
                                           rotation-velocity
                                           delta-t))))
         (new-x (saturate-x (+ old-x (* translation-quotient
                                        (cos new-angle)
                                        translation-velocity
                                        delta-t))))
         (new-y (saturate-y (- old-y (* translation-quotient
                                        (sin new-angle)
                                        translation-velocity
                                        delta-t))))
         (eaten-pills (filter (lambda (pill) (eaten? new-x new-y pill))
                              pills))
         (eaten-pills-no (length eaten-pills))
         (old-score (player-score player))
         (new-score (+ old-score eaten-pills-no)))
    (player-set-angle! player new-angle)
    ;; Don't move if the player would collide with any other player:
    (unless (exists? (lambda (other-player)
                       (< (distance new-x new-y
                                    (player-x other-player) (player-y other-player))
                          2player-radius))
                     (without-by-identity player players))
      (player-set-x! player new-x)
      (player-set-y! player new-y))
    (player-set-score! player new-score)
    (set! pills (subtract-lists-by-identity pills eaten-pills))
    (add-pills! eaten-pills-no)))

;;; The current list of players, as an unsorted list:
(define players
  '())

;;; The current list of pills, as an unsorted list:
(define pills
  (map (lambda whatever (make-pill))
       (iota pills-no)))
