/* This file is part of guile-with-graphics, a graphic library for Guile

Copyright (C) 2008, 2009  Luca Saiu
Based on turtle, by Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include "color.h"
#include "debug.h"
#include <SDL.h>
#include <libguile.h>
#include "sdl.h"

/* Convert the float coordinates in our stile (floats in [0.0, 1.0]) into
   unsigned 8-bit components, which is one format used by SDL; results
   are written in the given locations: */
static void game_floats_color_to_8bit_components(const float float_r,
                                                 const float float_g,
                                                 const float float_b,
                                                 Uint8 *byte_r, Uint8 *byte_g, Uint8 *byte_b){
  *byte_r = (Uint8)(float_r * 255.0);
  *byte_g = (Uint8)(float_g * 255.0);
  *byte_b = (Uint8)(float_b * 255.0);
}

/* Convert the given 8-bit color components, as used by SDL, into a single 32-bit
   value, another format also used by SDL: */
static Uint32 game_color_as_8bit_components_to_color_as_uint32(const Uint8 r,
                                                               const Uint8 g,
                                                               const Uint8 b){
  return SDL_MapRGB(the_window->default_surface->format, r, g, b);
}

void game_complete_color(game_color_t color){
  /* Turn the native color into 8 bit components: */
  game_floats_color_to_8bit_components(color->red, color->blue, color->green,
                                       &(color->red_as_uint8),
                                       &(color->blue_as_uint8),
                                       &(color->green_as_uint8));
  
  /* Turn the 8 bit components into a 32 bit value: */
  color->color_as_uint32 =
    game_color_as_8bit_components_to_color_as_uint32(color->red_as_uint8,
                                                     color->green_as_uint8,
                                                     color->blue_as_uint8);
}

/* This macro automatically generates a global const struct containing the color
   components as a static variable (so that it's not visible to other compilation
   units), and a pointer to it.
   If you want to look at the generated C code, just run
     gcc -E graphics.c
   and look at the output. Ask me if you want to understand it but have problems;
   it's quite easy. */
#define COLOR(NAME, RED, GREEN, BLUE) \
  static struct game_color game_color_ ## NAME ## _struct = { RED, GREEN, BLUE, 0, 0, 0, 0 }; \
  const game_color_t game_color_ ## NAME = (const game_color_t)&game_color_ ## NAME ## _struct;

/* /\* Call the macro once per color, to generate color definitions: *\/ */
 COLOR(black,   0, 0, 0)
/* COLOR(white,   1, 1, 1) */
/* COLOR(red,     1, 0, 0) */
/* COLOR(green,   0, 1, 0) */
/* COLOR(blue,    0, 0, 1) */
/* COLOR(cyan,    0, 1, 1) */
/* COLOR(magenta, 1, 0, 1) */
/* COLOR(yellow,  1, 1, 0) */

void game_initialize_colors(void){
  game_complete_color(game_color_black);
  /* game_complete_color(game_color_white); */
  /* game_complete_color(game_color_red); */
  /* game_complete_color(game_color_green); */
  /* game_complete_color(game_color_blue); */
  /* game_complete_color(game_color_cyan); */
  /* game_complete_color(game_color_magenta); */
  /* game_complete_color(game_color_yellow); */
}

/* The SMOB type of colors: */
static scm_t_bits game_color_smob_tag;

SCM game_mark_color(SCM color_smob){
  /* Do nothing: a color doesn't point to any heap-allocated object, so
     marking is trivial: */
  return SCM_BOOL_F; /* We have nothing else to mark: return an unboxed
                        object */
}

game_color_t game_color_smob_to_color(SCM color_smob){
  /* Check that the given SCM object is really a color SMOB: */
  scm_assert_smob_type(game_color_smob_tag, color_smob);
  
  /* Extract a pointer to the C struct from the SMOB: */
  game_color_t color = (game_color_t)SCM_SMOB_DATA(color_smob);
  return color;
}

SCM game_color_to_color_smob(game_color_t color){
  /* Create and return a SMOB for the given object: */
  SCM smob;
  SCM_NEWSMOB(smob, game_color_smob_tag, color);
  return smob;
}

static size_t game_free_color(SCM color_smob){
  /* Extract a pointer to the C struct from the SMOB: */
  game_color_t color = game_color_smob_to_color(color_smob);
  
  /* Destroying a color is easy; it's just a single buffer which
     contains no heap pointers (in fact it contains no pointers at
     all): */
  scm_gc_free(color, sizeof(struct game_color), "color");
  return 0; /* This *must* return 0 (see the Guile documenatation). */
}

static int game_print_color(SCM color_smob, SCM port, scm_print_state *pstate){
  /* Extract a pointer to the C struct from the SMOB: */
  game_color_t color = game_color_smob_to_color(color_smob);
  
  /* Print something reasonable: */
  char buffer[100];
  scm_puts ("#<color ", port);
  sprintf(buffer, "%f ", color->red); scm_puts (buffer, port);
  sprintf(buffer, "%f ", color->green); scm_puts (buffer, port);
  sprintf(buffer, "%f", color->blue); scm_puts (buffer, port);
  scm_puts (">", port);

  /* Non-zero means success: */
  return 1;
}

static SCM game_equalp_color(SCM color_smob1, SCM color_smob2){
  /* Extract pointers to the C structs from the SMOBs: */
  game_color_t color1 = game_color_smob_to_color(color_smob1);
  game_color_t color2 = game_color_smob_to_color(color_smob2);
  
  /* Two colors are equal only if they have the same RGB components: */
  if((color1->red == color2->red) &&
     (color1->green == color2->green) &&
     (color1->blue == color2->blue))
    return SCM_BOOL_T;
  else
    return SCM_BOOL_F;
}

/* Make a new color and return it as a color SMOB, so that it is usable from
   the Scheme side: */
SCM game_make_color_guile(SCM red, SCM green, SCM blue){
  /* Allocate space for the C structure in Guile's garbage-collected
     heap (the structure will be automatically destroyed when not needed
     any longer; this is why we *don't* use malloc() here): */
  game_color_t color =
    (game_color_t)scm_gc_malloc(sizeof(struct game_color), "color");
  
  /* Initialize the structure: */
  color->red = scm_to_double(red);
  color->green = scm_to_double(green);
  color->blue = scm_to_double(blue);
  if((color->red < 0.0) || (color->red > 1.0))
    scm_out_of_range("make-color", red);
  if((color->green < 0.0) || (color->green > 1.0))
    scm_out_of_range("make-color", green);
  if((color->blue < 0.0) || (color->blue > 1.0))
    scm_out_of_range("make-color", blue);
  game_complete_color(color);
  
  /* Create and return a SMOB for our new object: */
  return game_color_to_color_smob(color);
}

/* Register our new SMOB type for colors with Guile. This must be
   called just once at initialization time: */
void game_define_color_smob(void){
  game_color_smob_tag =
    scm_make_smob_type("color", sizeof(struct game_color));
  scm_set_smob_mark(game_color_smob_tag, game_mark_color);
  scm_set_smob_free(game_color_smob_tag, game_free_color);
  scm_set_smob_print(game_color_smob_tag, game_print_color);
  scm_set_smob_equalp(game_color_smob_tag, game_equalp_color);
}
